/*
 * Intel 80386 processor & "as" Assembel description
 *
 * In Linux, this file is to be compiled without any definitions.
 * In Windows & CYGWIN packet, define the variable "CYGWIN"
 * in "Makefile", or use the "Makefile.win".
 *
 * Difference between Linux and Windos+CYGWIN:
 *     - in Windows, the global names begin with underscore "_";
 *     - some assembler directives like ".type", ".size" do not work.
 */
#include <string.h>

#include "rtl.h"
#include "genasm.h"
#include "i80386.h"

const char * const REGISTER_NAME[8] = {
    "%eax", "%ebx", "%ecx", "%edx", "%edi", "%esi", "%ebp", "%esp"
};
const char * const BYTE_REGISTER_NAME[4] = {
    "%al", "%bl", "%cl", "%dl"
};
const int REGISTER_ALLOCATION_ORDER[NUM_REGISTERS_ALLOCATE] = {
    EAX, EDX, ECX, EBX, EDI, ESI
};

/* Set of instructions */
const instr_def instr_set[] = {
    {                           /* 0: move to register */
        RTX_SET, 2,
        ARG_RTX_ANY, 0,         /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_REGISTER, 0,    /* Result */

        2,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 = source */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n"
    },
    {                           /* 1: move from register */
        RTX_SET, 2,
        ARG_RTX_REGISTER, 0,    /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_ANY, 0,         /* Result */

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 = source */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_ANY, 0,             /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n"
    },
    {
        RTX_SET, 2,             /* 2: move memory byte -> register */
        ARG_RTX_BYTE_MEMORY, 0, /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_REGISTER, 0,    /* Result */

        2,                      /* Number of args */
        ARG_BYTE_MEMORY, 0,     /* arg0 = source */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovsbl\t;0,;2\n"
    },
    {                           /* 3: move register -> memory byte */
        RTX_SET, 2,
        ARG_RTX_REGISTER, 0,    /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_BYTE_MEMORY, 0, /* Result */

        2,                      /* Number of args */
        ARG_BYTE_REGISTER, BYTE_REGISTERS_SET, /* arg0 = source */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_BYTE_MEMORY, 0,     /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovb\t;0,;2\n"
    },
    {                           /* 4: push */
        RTX_SET, 2,
        ARG_RTX_ANY, 0,         /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_MEMREG_PREDEC, SP, /* Result */

        1,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 = source */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_NONE, 0,            /* result not used */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tpushl\t;0\n"
    },
    {                           /* 5: pop */
        RTX_SET, 2,
        ARG_RTX_MEMREG_POSTINC, SP, /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_ANY, 0,         /* Result */

        1,                      /* Number of args */
        ARG_NONE, 0,            /* arg0 not used */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_ANY, 0,             /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tpopl\t;0\n"
    },
    {                           /* 6: call */
        RTX_CALL, 2,
        ARG_RTX_ANY, 0,         /* Source */
        ARG_RTX_NONE, 0,
        ARG_RTX_HARDREG, EAX,   /* Result is R0 */

        1,                      /* Number of args */
        ARG_NAME, 0,            /* arg0 not used */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_NONE, 0,            /* result */
        (EAX_BIT | EDX_BIT | ECX_BIT),  /* Scratch registers */
        RES_SEPARATE,
        "\tcall\t;0\n"
    },
    {
        RTX_PLUS, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_REGISTER, 0,        /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t(;0,;1),;2\n"
    },
    {
        RTX_PLUS, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY_CONST, 0,   /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_PLAIN_CONST, 0,     /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t;1(;0),;2\n"
    },
    {
        RTX_PLUS, 3,
        ARG_RTX_ANY_CONST, 0,   /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_PLAIN_CONST, 0,     /* arg0 */
        ARG_REGISTER, 0,        /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t;0(;1),;2\n"
    },
    {
        RTX_PLUS, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\taddl\t;1,;2\n"
    },
    {
        RTX_MINUS, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tsubl\t;1,;0\n"
    },
    {
        RTX_MINUS, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\tsubl\t;1,;2\n"
    },
    {
        RTX_MUL, 3,             /* res = arg0*2; */
        ARG_RTX_REGISTER, 0,
        ARG_RTX_CONST, 2,
        ARG_RTX_REGISTER, 0,

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t0(,;0,2),;2\n"
    },
    {
        RTX_MUL, 3,             /* res = 2*arg0; */
        ARG_RTX_CONST, 2,
        ARG_RTX_REGISTER, 0,
        ARG_RTX_REGISTER, 0,

        2,                      /* Number of args */
        ARG_NONE, 0,            /* arg0 not used */
        ARG_REGISTER, 0,        /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t0(,;1,2),;2\n"
    },
    {
        RTX_MUL, 3,             /* res = arg0*4; */
        ARG_RTX_REGISTER, 0,
        ARG_RTX_CONST, 4,
        ARG_RTX_REGISTER, 0,

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_NONE, 0,            /* arg1 not used */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t0(,;0,4),;2\n"
    },
    {
        RTX_MUL, 3,             /* res = 4*arg0; */
        ARG_RTX_CONST, 4,
        ARG_RTX_REGISTER, 0,
        ARG_RTX_REGISTER, 0,

        2,                      /* Number of args */
        ARG_NONE, 0,            /* arg0 not used */
        ARG_REGISTER, 0,        /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tleal\t0(,;1,4),;2\n"
    },
    {
        RTX_MUL, 3,             /* res = arg0*arg1; */
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result in any register */
        0,                      /* Scratch registers */
        RES_IN_ARG0,            /* Result in arg0 */
        "\timul\t;1,;0\n"
    },
    {
        RTX_MUL, 3,             /* res = arg0*arg1; */
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result in any register */
        0,                      /* Scratch registers */
        RES_SEPARATE,           /* Result in arg0 */
        "\tmovl\t;0,;2\n\timul\t;1,;2\n"
    },
    {
        RTX_DIV, 3,             /* res = arg0/arg1; */
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_REGISTER_SET, EAX_BIT, /* arg0 in %eax */
        ARG_REGISTER, 0,           /* arg1 */
        ARG_REGISTER, 0,        /* result in %eax, in arg0 */
        (EAX_BIT | EDX_BIT),    /* Scratch registers */
        RES_IN_ARG0,            /* Result in arg0 */
        "\tcltd\n\tidivl\t;1\n"
    },
    {
        RTX_MOD, 3,             /* res = arg0%arg1; */
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_REGISTER_SET, EAX_BIT, /* arg0 in %eax */
        ARG_REGISTER, 0,           /* arg1 */
        ARG_REGISTER_SET, EDX_BIT, /* result in %edx */
        (EAX_BIT | EDX_BIT),    /* Scratch registers */
        RES_SEPARATE,
        "\tcltd\n\tidivl\t;1\n"
    },
    {
        RTX_DIV, 3,             /* res = arg0/arg1; */
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_REGISTER, 0,        /* arg1 */
        ARG_REGISTER_SET, EAX_BIT, /* result in %eax */
        (EAX_BIT | EDX_BIT),    /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\tcltd\n\tidivl\t;1\n"
    },
    {
        RTX_MOD, 3,             /* res = arg0/arg1; */
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER_SET, EDX_BIT, /* result in %edx */
        (EAX_BIT | EDX_BIT),    /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,%eax\n\tcltd\n\tidivl\t;1\n"
    },
    {
        RTX_DIV, 3,             /* res = arg0/arg1; */
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY_CONST, 0,   /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_REGISTER_SET, EAX_BIT,      /* arg0 */
        ARG_CONST, 0,                   /* arg1 we copy it to %ecx */
        ARG_REGISTER, 0,                /* result in %eax */
        (EAX_BIT | EDX_BIT | ECX_BIT),  /* Scratch registers */
        RES_IN_ARG0,
        "\tmovl\t;1,%ecx\n\tcltd\n\tidivl\t%ecx\n"
    },
    {
        RTX_MOD, 3,             /* res = arg0/arg1; */
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY_CONST, 0,   /* arg1 */
        ARG_RTX_REGISTER, 0,    /* res */

        3,                      /* Number of args */
        ARG_REGISTER_SET, EAX_BIT,      /* arg0 */
        ARG_CONST, 0,                   /* arg1 we copy it to %ecx */
        ARG_REGISTER_SET, EDX_BIT,      /* result in %edx */
        (EAX_BIT | EDX_BIT | ECX_BIT),  /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;1,%ecx\n\tcltd\n\tidivl\t%ecx\n"
    },
    {
        RTX_AND, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tandl\t;1,;0\n"
    },
    {
        RTX_AND, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\tandl\t;1,;2\n"
    },
    {
        RTX_OR, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\torl\t;1,;0\n"
    },
    {
        RTX_OR, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\torl\t;1,;2\n"
    },
    {
        RTX_XOR, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\txorl\t;1,;0\n"
    },
    {
        RTX_XOR, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tmovl\t;0,;2\n\txorl\t;1,;2\n"
    },
    {
        RTX_CMP, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_CC0, 0,         /* Result */

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_NONE, 0,            /* result */
        0,                      /* Scratch registers */
        RES_SEPARATE,
        "\tcmpl\t;1,;0\n"
    },
    {
        RTX_CMP, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_ANY, 0,         /* arg1 */
        ARG_RTX_CC0, 0,         /* Result */

        2,                      /* Number of args */
        ARG_ANY, 0,             /* arg0 */
        ARG_ANY, 0,             /* arg1 */
        ARG_NONE, 0,            /* result */
        EAX_BIT,                /* Scratch register: %eax */
        RES_SEPARATE,
        "\tmovl\t;0,%eax\n\tcmpl\t;1,%eax\n"
    },
    {
        RTX_LSHIFT, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_REGISTER_SET, ECX_BIT, /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tsall\t%cl,;0\n"
    },
    {
        RTX_LSHIFT, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY_CONST, 0,   /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_CONST, 0,           /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tsall\t;1,;0\n"
    },
    {
        RTX_LSHIFT, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_REGISTER_SET, ECX_BIT, /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tmovl;0,;2\n\tsall\t%cl,;2\n"
    },
    {
        RTX_RSHIFT, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_REGISTER_SET, ECX_BIT, /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tsarl\t%cl,;0\n"
    },
    {
        RTX_RSHIFT, 3,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_ANY_CONST, 0,   /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_CONST, 0,           /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tsarl\t;1,;0\n"
    },
    {
        RTX_RSHIFT, 3,
        ARG_RTX_ANY, 0,         /* arg0 */
        ARG_RTX_REGISTER, 0,    /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        3,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_REGISTER_SET, ECX_BIT, /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tmovl;0,;2\n\tsarl\t%cl,;2\n"
    },
    {
        RTX_UNMINUS, 2,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_NONE, 0,        /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_NONE, 0,            /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tnegl\t;0\n"
    },
    {
        RTX_NOT, 2,
        ARG_RTX_REGISTER, 0,    /* arg0 */
        ARG_RTX_NONE, 0,        /* arg1 */
        ARG_RTX_REGISTER, 0,    /* Result */

        2,                      /* Number of args */
        ARG_REGISTER, 0,        /* arg0 */
        ARG_NONE, 0,            /* arg1 */
        ARG_REGISTER, 0,        /* result */
        0,                      /* Scratch registers */
        RES_IN_ARG0,
        "\tnotl\t;0\n"
    },


    {
        RTX_NONE
    }
};

/* Some fixed instruction that must be defined */
const instr_def_ptr INSTR_MOV_ANY_REG = (const instr_def_ptr) &(instr_set[0]);
const instr_def_ptr INSTR_MOV_REG_ANY = (const instr_def_ptr) &(instr_set[1]);
const instr_def_ptr INSTR_MOV_BYTE_REG = (const instr_def_ptr) &(instr_set[2]);
const instr_def_ptr INSTR_MOV_REG_BYTE = (const instr_def_ptr) &(instr_set[3]);
const instr_def_ptr INSTR_PUSH = (const instr_def_ptr) &(instr_set[4]);
const instr_def_ptr INSTR_POP = (const instr_def_ptr) &(instr_set[5]);
const instr_def_ptr INSTR_CALL = (const instr_def_ptr) &(instr_set[6]);

const char * const file_preambula[] = {
    ".file\t\";0\"\n",          /* ";0" is a template for file name */
    "\t.version\t\"01.01\"\n",
    "tinycc_compiled.:\n",
    ".text\n",
    NULL
};

const char * const file_postambula[] = {
    "\t.ident\t\"TINYCC: 1.0.0 20020114 (release)\"\n",
    NULL
};

#ifndef __CYGWIN__
/* Linux */
const char * const function_preambula[] = {
    "\t.align\t4\n",
    ".globl\t;0\n",     /* ";0" is a template for function name */
    "\t.type\t;0,@function\n",
    ";0:\n",
    "\tpushl\t%ebp\n",
    "\tmovl\t%esp,%ebp\n",
    NULL
};
#else
/* Windows & Cygwin */
const char * const function_preambula[] = {
    "\t.align\t4\n",
    ".globl\t_;0\n",     /* ";0" is a template for function name */
    /*... "\t.type\t_;0,@function\n", ...*/
    "_;0:\n",
    "\tpushl\t%ebp\n",
    "\tmovl\t%esp,%ebp\n",
    NULL
};
#endif  /* __CYGWIN__ */

#ifndef __CYGWIN__                      /* Linux */
const char * const function_postambula[] = {
    "\tleave\n",
    "\tret\n",
    ";0:\n",                /* ";0" is a template for function end label */
    "\t.size\t;1,;0-;1\n",  /* ";1" is a function name */
    NULL
};
#else                                   /* Windows & CYGWIN */
const char * const function_postambula[] = {
    "\tleave\n",
    "\tret\n",
    ";0:\n",                /* ";0" is a template for function end label */
    /*... "\t.size\t_;1,;0-;1\n", ...*/ /* ";1" is a function name */
    NULL
};
#endif  /* __CYGWIN__ */

const char * const stack_advance_instruction = "\tsubl\t$;0,%esp\n";
const char * const label_line = ";0:\n";
const char * const goto_instruction = "\tjmp\t;0\n";
const char * const branch_instruction = "\tj;1\t;0\n";

/* The order of conditions is fixed!
   ==, !=, >, >=, <, <=,
   unsigned >, >=, <, <= */
const char * const asm_condition_names[] = {
    "e", "ne", "g", "ge", "l", "le",
    "a", "ae", "b", "be", NULL
};

/* Arguments of instructions */
const char * const arg_const = "$;0";               /* $10 */

#ifndef __CYGWIN__                                  /* Linux: */
const char * const arg_name = "$;0";                /* $xyz */
const char * const arg_global_name = "$;0";         /* $xyz */
const char * const arg_mem_name = ";0";             /* xyz */
const char * const arg_mem_global_name = ";0";      /* xyz */
#else                                               /* Windows: */
const char * const arg_name = "$;0";                /* $xyz */
const char * const arg_global_name = "$_;0";        /* $_xyz */
const char * const arg_mem_name = ";0";             /* xyz */
const char * const arg_mem_global_name = "_;0";     /* _xyz */
#endif  /* CYGWIN */

const char * const arg_mem_reg = "(;0)";            /* (%esi) */
const char * const arg_mem_reg_const = ";1(;0)";    /* 12(%ebp) */

/* Templates for global definitions */
#ifndef __CYGWIN__                                  /* Linux: */
const char * const gdef_int = "\t.comm\t;0,4,4\n";
const char * const gdef_int_array = "\t.comm\t;0,;1,32\n";
const char * const gdef_char = "\t.comm\t;0,1,1\n";;
const char * const gdef_char_array = "\t.comm\t;0,;1,1\n";
#else                                               /* Windows: */
const char * const gdef_int = "\t.comm\t_;0,4,4\n";
const char * const gdef_int_array = "\t.comm\t_;0,;1,32\n";
const char * const gdef_char = "\t.comm\t_;0,1,1\n";;
const char * const gdef_char_array = "\t.comm\t_;0,;1,1\n";
#endif  /* __CYGWIN__ */

const char * const gdef_extern_name = NULL; /* Not used in this assembler */

const char * const gdef_preambula = "\n";
const char * const gdef_postambula = NULL;   /* Not used */

/* Templates for string definitions */
const char * const strdef_preambula = ".section\t.rodata\n";
const char * const strdef_postambula = NULL;

const char * const strdef_string = ";0:\n\t.string\t\";1\"\n";

