#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rtl.h"
#include "scanner.h"
#include "comp.h"

block_header rtl_function;      /* rtl of current function */

rtx_ptr allocate_rtx() {
    rtx_ptr e = (rtx_ptr) malloc(sizeof(rtx));
    memset(e, 0, sizeof(rtx));
    return e;
}

rtx_ptr clone_rtx(rtx_ptr e) {
    rtx_ptr s = allocate_rtx();
    *s = *e;
    if (e->left != NULL)
        s->left = clone_rtx(e->left);
    if (e->right != NULL)
        s->right = clone_rtx(e->right);
    return s;
}

void release_rtx(rtx_ptr e) {
    if (e == NULL) return;
    if (e->left != NULL)
        release_rtx(e->left);
    if (e->right != NULL)
        release_rtx(e->right);
    free(e);
}

void init_rtl() {
    lst_link(&rtl_function, &rtl_function);
}

rtl_ptr allocate_rtl_line() {
    rtl_ptr l = (rtl_ptr) malloc(sizeof(rtl));
    memset(l, 0, sizeof(rtl));
    return l;
}

void release_rtl_line(rtl_ptr l) {
    if (l->expr != NULL)
        release_rtx(l->expr);
    free(l);
}

void release_rtl_function() {
    block_ptr l = rtl_function.next;
    while (l != &rtl_function) {
        lst_remove(l);
        release_rtl_line((rtl_ptr) l);
        l = rtl_function.next;
    }
}

void add_rtl_line(rtl_ptr line) {
    lst_link(rtl_function.prev, &(line->link));
    lst_link(&(line->link), &rtl_function);
}

void add_rtl_function_beg() {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_FUNCTION_BEG;
    add_rtl_line(l);
}

void add_rtl_function_end() {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_FUNCTION_END;
    add_rtl_line(l);
}

void add_rtl_label(int lab) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_LABEL;
    l->int_value = lab;
    add_rtl_line(l);
}

void add_rtl_goto(int lab) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_GOTO;
    l->int_value = lab;

    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;

    l->expr->left = allocate_rtx();
    l->expr->left->oper = RTX_HARDREG;
    l->expr->left->hard_reg = PC;

    l->expr->right = allocate_rtx();
    l->expr->right->oper = RTX_LABEL;
    l->expr->right->int_value = lab;

    add_rtl_line(l);
}

void add_rtl_cond_goto(int condition, int lab) {
    rtl_ptr l = allocate_rtl_line();
    l->typ = RTL_GOTO;
    l->int_value = lab;
    l->int_value2 = condition;

    l->expr = allocate_rtx();
    l->expr->oper = RTX_SET;

    l->expr->left = allocate_rtx();
    l->expr->left->oper = RTX_HARDREG;
    l->expr->left->hard_reg = PC;

    l->expr->right = allocate_rtx();
    l->expr->right->oper = RTX_IF_THEN_ELSE;
    l->expr->right->int_value = condition;

    l->expr->right->left = allocate_rtx();
    l->expr->right->left->oper = RTX_LABEL;
    l->expr->right->left->int_value = lab;

    l->expr->right->right = allocate_rtx();
    l->expr->right->right->oper = RTX_HARDREG;
    l->expr->left->hard_reg = PC;

    add_rtl_line(l);
}

static const char *RTL_LINE_TYPES[] = {
    "???",
    "RTL_FUNCTION_BEG",      /* 1 */
    "RTL_FUNCTION_END",      /* 2 */
    "RTL_LABEL",             /* 3 */
    "RTL_GOTO",              /* 4 */
    "RTL_USE",               /* 5 */
    "RTL_INSTR",             /* 6 */
    ""
};

static int rtx_level = 0;

static int print_rtl_line(block_ptr line, void *arg0, long arg1) {
    rtl_ptr l = (rtl_ptr) line;
    fprintf(dump_file, "%s ", RTL_LINE_TYPES[l->typ]);
    if (l->typ == RTL_LABEL || l->typ == RTL_GOTO || l->typ == RTL_USE) {
        fprintf(dump_file, "%d ", l->int_value);
        if (l->typ == RTL_GOTO && l->int_value2 != 0)
            fprintf(dump_file, "if %s", cond_name(l->int_value2)); /* Conditional goto */
    } else if (l->typ == RTL_INSTR) {
        rtx_level = 0;
        print_rtx(l->expr);
    }
    fprintf(dump_file, "\n");
    return 0;
}

void print_rtl_function() {
    if (!(compiler_flags & DUMP_RTL))
        return;
    lst_loop(&rtl_function, &print_rtl_line, NULL, 0);
}

static const char *RTX_NAMES[] = {
    "???",
    "RTX_REG",          /* 1 */
    "RTX_HARDREG",      /* 2 */
    "RTX_MEM",          /* 3 */
    "RTX_NAME",         /* 4 */
    "RTX_ICON",         /* 5 */
    "RTX_SET",          /* 6 */
    "RTX_PLUS",         /* 7 */
    "RTX_MINUS",        /* 8 */
    "RTX_MUL",          /* 9 */
    "RTX_DIV",          /* 10 */
    "RTX_MOD",          /* 11 */
    "RTX_AND",          /* 12 */
    "RTX_OR",           /* 13 */
    "RTX_XOR",          /* 14 */
    "RTX_CMP",          /* 15 */
    "RTX_TST",          /* 16 */
    "RTX_LSHIFT",       /* 17 */
    "RTX_RSHIFT",       /* 18 */
    "RTX_INC",          /* 19 */
    "RTX_DEC",          /* 20 */
    "RTX_CALL",         /* 21 */
    "RTX_LABEL",        /* 22 */
    "RTX_IF_THEN_ELSE", /* 23 */
    "RTX_PREDEC",       /* 24 */
    "RTX_POSTINC",      /* 25 */
    "RTX_UNMINUS",      /* 26 */
    "RTX_NOT",          /* 27 */
    ""
};

void print_rtx(rtx_ptr e) {
    int op;
    int i;
    if (e == NULL) return;

    for (i = 0; i < rtx_level; i++)
        fprintf(dump_file, "    ");     /* line indent */

    op = e->oper;
    fprintf(dump_file, "%s", RTX_NAMES[op]);
    switch (op) {
        case RTX_REG:
            fprintf(dump_file, "[%d]", e->reg); return;
        case RTX_HARDREG:
            if (e->hard_reg == FP) fprintf(dump_file, " FP");
            else if (e->hard_reg == SP) fprintf(dump_file, " SP");
            else if (e->hard_reg == PC) fprintf(dump_file, " PC");
            else if (e->hard_reg == CC0) fprintf(dump_file, " CC0");
            else fprintf(dump_file, "[%d]", e->hard_reg);
            return;
        case RTX_ICON:
        case RTX_LABEL:
            fprintf(dump_file, " %d", e->int_value); return;
        case RTX_NAME:
            fprintf(dump_file, " %s", e->string_value); return;
        case RTX_MEM:
        case RTX_PREDEC:
        case RTX_POSTINC:
            if (e->modifiers & RTX_MODBYTE)
                fprintf(dump_file, ":B");
            break;
    }
    if (e->left != NULL) {
        fprintf(dump_file, "\n"); 
        rtx_level++; print_rtx(e->left); rtx_level--;
    }
    if (e->right != NULL) {
        fprintf(dump_file, "\n"); 
        rtx_level++; print_rtx(e->right); rtx_level--;
    }
}

const char *cond_name(int condition) {
    const char *n = "??";
    if (condition == EQ)      n = "eq";
    else if (condition == NE) n = "ne";
    else if (condition == GT) n = "gt";
    else if (condition == GE) n = "ge";
    else if (condition == LT) n = "lt";
    else if (condition == LE) n = "le";
    else if (condition == GTU) n = "gtu";
    else if (condition == GEU) n = "geu";
    else if (condition == LTU) n = "ltu";
    else if (condition == LEU) n = "leu";
    return n;
}

int inverse_cond(int condition) {
    int ic = 0;
    if (condition == EQ)      ic = NE;
    else if (condition == NE) ic = EQ;
    else if (condition == GT) ic = LE;
    else if (condition == GE) ic = LT;
    else if (condition == LT) ic = GE;
    else if (condition == LE) ic = GT;
    else if (condition == GTU) ic = LEU;
    else if (condition == GEU) ic = LTU;
    else if (condition == LTU) ic = GEU;
    else if (condition == LEU) ic = GTU;
    return ic;
}

int rtx_inverse_cond(int tree_cond);
