%{
/* Test scanner for C */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define NAME     1
#define INTCONST 2
#define IF       3
#define ELSE     4
#define WHILE    5

#define NAME_MAXLEN 128

const char* TOKEN_NAMES[] = {
    "EOF",
    "NAME",
    "INTCONST",
    "IF",
    "ELSE",
    "WHILE"
};

typedef struct {
    char text_value[NAME_MAXLEN];
    int int_value;
} YYSTYPE;

YYSTYPE yylval;

%}

ALPHA           [A-Za-z]
DIGIT           [0-9]
ALPHADIGIT      ({ALPHA}|{DIGIT})

%Start OUT CCOMMENT CPPCOMMENT

%%
                        { BEGIN OUT; }
<OUT>"if"               { return IF; }
<OUT>"else"             { return ELSE; }
<OUT>"while"            { return WHILE; }
<OUT>{ALPHA}{ALPHADIGIT}* {
                            strncpy(yylval.text_value, yytext, NAME_MAXLEN);
                            return NAME;
                        }
<OUT>{DIGIT}+           {
                            yylval.int_value = atoi(yytext);
                            return INTCONST;
                        }
<OUT>"/*"               { BEGIN CCOMMENT; }
<CCOMMENT>(.|\n)        {}
<CCOMMENT>"*/"          { BEGIN OUT; }
<OUT>"//"               { BEGIN CPPCOMMENT; }
<CPPCOMMENT>.           {}
<CPPCOMMENT>\n          { BEGIN OUT; }
<OUT>.                  {}
<OUT>\n                 {}

%%

int main() {
    int token;
    while ((token = yylex()) != 0) {
        printf("Lexeme %s", TOKEN_NAMES[token]);
        if (token == NAME) {
            printf(" value=%s", yylval.text_value);
        } else if (token == INTCONST) {
            printf(" value=%d", yylval.int_value);
        }
        printf("\n");
    }
}

int yywrap() { return 1; }
