/* 
 * String pool
 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "strpool.h"
#include "scanner.h"

static const char * const stringPrefix = STRING_PREFIX;
static char stringNameBuffer[STRMAX + 1];
static int stringID = 0;

block_header string_list;

void init_string_list() {
    lst_link(&string_list, &string_list);
}

static int string_cmp(block_ptr s, void *str, long dummy) {
    if (strcmp(((string_ptr) s)->str, (const char *) str) == 0)
        return 1;
    return 0;
}

string_ptr lookupStringEntry(const char *str) {
    return (string_ptr) lst_loop(
        &string_list, &string_cmp, (void*) str, (long) 0
    );
}

const char *lookupString(const char *str) {
    string_ptr s = lookupStringEntry(str);
    if (s == NULL)
        return NULL;
    else
        return s->str;
}

static int string_id_cmp(block_ptr s, void *str, long id) {
    if (((string_ptr) s)->str_id == id)
        return 1;
    return 0;
}

string_ptr lookupStringIdEntry(int id) {
    return (string_ptr) lst_loop(
        &string_list, &string_id_cmp, NULL, (long) id
    );
}

const char *lookupStringId(int id) {
    string_ptr s = lookupStringIdEntry(id);
    if (s == NULL)
        return NULL;
    else
        return s->str;
}

string_ptr addString(const char *str) {
    int l = strlen(str);
    string_ptr s = lookupStringEntry(str);
    if (s == NULL) {
        /* Add a string to the string pool */
        s = allocate_string_entry();
        s->str = (char *) malloc(l + 1);
        strcpy(s->str, str);
        lst_link(&(s->link), string_list.next);
        lst_link(&string_list, &(s->link));
    }
    return s;
}

string_ptr allocate_string_entry() {
    string_ptr s = (string_ptr) malloc(sizeof(string_entry));
    memset(s, 0, sizeof(string_entry));
    s->str_id = ++stringID;
    strncpy(s->str_name, getStringName(s->str_id), MAX_STR_NAME_LENGTH);
    s->str_name[MAX_STR_NAME_LENGTH-1] = 0;
    return s;
}

void release_string_entry(string_ptr s) { 
    if (s->str != NULL)
        free(s->str);
    free(s);
}

void release_string_list() {
    block_ptr s = string_list.next;
    while (s != &string_list) {
        lst_remove(s);
        release_string_entry((string_ptr) s);
        s = string_list.next;
    }
}

const char *getStringNamePrefix() {
    return stringPrefix;
}

const char *getStringName(int strId) {
    sprintf(stringNameBuffer, "%s%d", getStringNamePrefix(), strId);
    return stringNameBuffer;
}
