#ifndef TREE_H
#define TREE_H

#include "flagbits.h"
#include "list.h"
#include "nametab.h"
#include "strpool.h"

struct tree_tag;

typedef struct tree_tag {
    block_header link;
    struct tree_tag* left;
    struct tree_tag* right;
    int oper;           /* NAME, PLUS, MUL, PNT, etc */
    name_ptr name;      /* reference into a name table */
    int int_value;           /* integer constant, value of relation */
    string_ptr string_value; /* pointer to a string for string constant */
    int modifiers;      /* various flags */
    int reg;            /* pseudo register */
    int base_type;      /* expr. base type: BASETYPE_INT, ... (-1) incorrect */
    int pointer;        /* 0 base type, 1 pointer, 2 pnt. to pnt, ... */
} tree, *tree_ptr;

/* Flags in the "modifiers" field of struct tree_tag 
   (defined in "flagbits.h"):
MOD_POSTFIX     Postfix operation ++, --
MOD_BYTE        byte (character) operation
MOD_BOOL        boolean expression
MOD_LVAL        l-value
*/

/* Node operations: may be either tokens, such as NAME, PLUS, etc.,
   or the following values: */
#define NODE_TYPE_BASE 65536
#define PNT     (NODE_TYPE_BASE+1)  /* pointer */
#define ARR     (NODE_TYPE_BASE+2)  /* array element */
#define FUN     (NODE_TYPE_BASE+3)  /* function call */
#define UNMINUS (NODE_TYPE_BASE+4)  /* unary minus */
#define AMP     (NODE_TYPE_BASE+5)  /* & (address) */
#define LST     (NODE_TYPE_BASE+6)  /* list of function arguments */

void init_trees();
void release_tree(tree_ptr t);
void release_tree_node(tree_ptr n);
tree_ptr allocate_tree_node();

/* The list of tree nodes allocated (used for releasing of trees) */
extern block_header tree_nodes_allocated;
void release_all_trees();

/* debug printing */
void print_tree(tree*);

int expr_type(tree_ptr e);  /* Returns 0, if correct, (-1) if error */
                            /* Fills in recursively e->base_type, e->pointer */

#endif  /* TREE_H */
